/*

Licensed under the GNU AFFERO GENERAL PUBLIC LICENSE Version 3 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package application

import (
	"strings"

	settings "k8s.libre.sh/controller-utils/application/settings"
	"k8s.libre.sh/controller-utils/application/settings/parameters"
	"k8s.libre.sh/controller-utils/interfaces"
	"k8s.libre.sh/controller-utils/meta"
	"k8s.libre.sh/controller-utils/status"
)

type Instance interface {
	meta.Instance
	GetComponents() map[string]Component
	GetJobs() map[string]Component
	GetOwner() status.ObjectWithStatus
	GetSettings() map[string]settings.Component
	GetApplicationStatus() *status.ApplicationStatus
	SetDefaults()
}

func GetTemplateValues(i Instance, setts map[string]settings.Component) map[string]interface{} {
	values := make(map[string]interface{})
	cptsVals := make(map[string]interface{})
	settsVals := make(map[string]interface{})

	for _, cpt := range i.GetComponents() {
		// TODO create templateValues for Component interface
		cptsVals[cpt.GetMeta().GetComponent()] = cpt.(interface{})
	}

	for _, sett := range setts {
		// TODO create templateValues for Settings interface
		settsVals[sett.GetMeta().GetComponent()] = parameters.KeyPairValues(sett.GetConfig().GetParameters())
	}

	values["components"] = cptsVals
	values["settings"] = settsVals

	return values

}

func AppendIfUnique(objs []status.ObjectStatus, obj status.ObjectStatus) []status.ObjectStatus {

	if len(objs) == 0 {
		objs = append(objs, obj)
	}

	for _, o := range objs {
		if o.Name == obj.Name && o.Kind == obj.Kind {
			return objs
		}
	}
	objs = append(objs, obj)
	return objs
}

func GetWatchedObjectsFromParameters(ps *parameters.Parameters) status.Resources {

	watched := status.Resources{}

	for _, p := range *ps {
		if len(p.Ref) > 0 && p.Type != parameters.ObjectFieldParameter && p.RefType != parameters.ObjectFieldParameter {
			obj := status.ObjectStatus{
				Link:    "",
				Name:    p.Ref,
				Version: "",
				Kind:    "Secret",
				Group:   "",
			}
			watched.Objects = AppendIfUnique(watched.Objects, obj)
		}
	}

	return watched
}

func GetWatchedObjectsFromSources(srcs *settings.Sources) status.Resources {

	watched := status.Resources{}

	for _, src := range *srcs {
		obj := status.ObjectStatus{
			Link:    "",
			Name:    src.Ref,
			Version: "",
			Kind:    src.Type,
			Group:   "",
		}

		watched.Objects = AppendIfUnique(watched.Objects, obj)
	}

	return watched
}

func InitSettings(i Instance, r interfaces.Reconcile) error {

	appStatus := i.GetApplicationStatus()
	setts := i.GetSettings()

	watched := status.Resources{}

	// First we init external values and random values
	for name, sett := range setts {

		watched.Objects = append(watched.Objects, GetWatchedObjectsFromParameters(sett.GetParameters()).Objects...)
		watched.Objects = append(watched.Objects, GetWatchedObjectsFromSources(sett.GetSources()).Objects...)
		appStatus.Watched = watched

		if len(appStatus.Settings[name].Resources.Objects) > 0 {
			for _, o := range appStatus.Settings[name].Resources.Objects {
				src := settings.Source{
					Ref: o.Name,
					// & TO FIX
					Type: strings.ToLower(o.Kind),
				}
				// TODO append if unique
				*sett.GetConfig().GetSources() = append(*sett.GetConfig().GetSources(), src)
			}
		}

		err := settings.Init(sett, r.GetClient(), i.GetOwner(), r.GetScheme())

		if err != nil {
			return err
		}
	}

	// Transform instance to template values used to init settings
	values := GetTemplateValues(i, i.GetSettings())

	// Init the values from templates & set defaults
	for _, sett := range setts {
		err := sett.GetConfig().GetParameters().InitTemplateValues(values)
		if err != nil {
			return err
		}

		sett.SetDefaults()
	}

	return nil
}

// Initiatilize & Defaults
func Init(i Instance, r interfaces.Reconcile) error {

	i.SetDefaults()

	err := InitSettings(i, r)

	if err != nil {
		return err
	}

	// Init Components
	cpts := i.GetComponents()
	InitComponentsFromSettings(cpts, i.GetSettings())

	return nil
}
